require 'uri'
require 'helix_web_services_client/open_model'

class HelixWebServicesClient

  #
  # Actions connected to managing user SSH keys for GitFusion
  #

  # Return all keys for specified user. Every entry contains key name and its contents.
  #
  # @param user [String] Username for which keys will be fetched
  def keys(user)
    path = "/git-fusion/v1/users/#{user}/keys"
    execute_method_no_body(:get, path)
  end

  # Add a key for a specified user. It will try to add a new SSH key for specified user under
  # a specific name. It will fail if the same key belongs to a different user,
  # if the same key is under a different name and if there is already a key under the specified name.
  # In case key with the same content is present for specified user under the same name - it will skip
  # re-submission.
  #
  # @param user [String] Username to assign SSH key to
  # @param key_name [String] File name to save the key under
  # @param key [String] Content of SSH key
  def add_key(user, key_name, key)
    path = "/git-fusion/v1/users/#{user}/keys/#{key_name}"
    execute_method_with_body(:put, path, {'key' => key})
  end

  # Delete a specific SSH key. It accepts an SSH key looks up the owner and then proceeds to remove it.
  #
  # @param key [String] SSH key string to be removed
  def delete_key(user, key_name)
    path = "/git-fusion/v1/users/#{user}/keys/#{key_name}"
    execute_method_no_body(:delete, path)
  end

  # Delete all keys for a user. All keys for specified user will be deleted from perforce.
  #
  # @param user [String] Chosen user to remove all keys for
  def delete_keys(user)
    path = "/git-fusion/v1/users/#{user}/keys"
    execute_method_no_body(:delete, path)
  end

  # URI encode component, to use in an endpoint while preserving formatting of a resource.
  # API will decode the string and proceed with action.

  # @param str [String] Component to be URI escaped
  def encode_component(str)
    str.chars.map do |ch|
      if GitFusionStrings::URI_RESERVED_CHARACTERS.include?(ch)
        '%' + ch.ord.to_s(16)
      else
        ch
      end
    end.join('')
  end
end