require_relative './test_config'

require 'helix_web_services_client'

describe 'HelixWebServicesClient Helix Sync' do

  my_project_id = 'My~20Project'

  context 'create_helix_sync_device_client' do
    it 'should create a new device client for a user' do
      client_as_jdoe do |c|
        c.add_setting('HVE_PROJECTS_PATH', '//depot/main')
        info = c.create_helix_sync_device_client(my_project_id, 'client_user_test', '/dev/null')

        expect(info.client).to include(my_project_id)
        expect(info.client).to include('client_user_test')
        expect(info.client).to include('jdoe')
      end
    end
  end

  context 'create_helix_sync_shelf_client' do
    it 'should create a new shelf client for a user' do
      client_as_jdoe do |c|
        c.add_setting('HVE_PROJECTS_PATH', '//depot/main')
        info = c.create_helix_sync_shelf_client(my_project_id)

        expect(info.client).to include(my_project_id)
        expect(info.client).to include('jdoe')
      end
    end
  end

  context 'fetch_helix_sync_latest_changelist' do
    it 'should return a latest changelist for a user' do
      client_as_jdoe do |c|
        c.add_setting('HVE_PROJECTS_PATH', '//depot/main')
        info = c.fetch_helix_sync_latest_changelist(my_project_id)

        expect(info.change.to_i).to be > 1
      end
    end
  end

  context 'preview_helix_sync_pending_change' do
    context 'simple add' do
      it 'indicates a new file should be added' do
        client_as_jdoe do |c|
          c.add_setting('HVE_PROJECTS_PATH', '//depot/main')

          begin
            p4 = p4_as_jdoe

            info = create_sync_project('//depot/main/preview_test_add', c, p4)
            sync_root = info[0]
            device_client = info[1]

            new_file_name = create_new_file_under(sync_root)

            run_device_sync('//depot/main/preview_test_add', sync_root, c, device_client)

            preview = c.preview_helix_sync_pending_change('preview_test_add')

            puts "preview: #{preview}"

            expect(preview.length).to eq(1)

            expect(preview.first.depotFile).to eq("//depot/main/preview_test_add/#{new_file_name}")
            expect(preview.first.action).to eq('add')

            # Cleanup
            c.delete_helix_sync_pending_changelist('preview_test_add')
            c.delete_helix_sync_device_client('preview_test_add', 'test-machine')
            c.delete_helix_sync_shelf_client('preview_test_add')

          ensure
            p4.disconnect if !p4.nil? and p4.connected?
            FileUtils.rmtree(sync_root) if sync_root && Dir.exist?(sync_root)
          end
        end
      end
    end

    # This is a scenario where we've deleted a file on a shelf, that was also
    # already deleted by someone else. So the "resolution" is to just remove
    # the file from the shelf, no further action is needed.
    context 'remove from shelf' do
      # This test depends on the file added above
      it 'just removes the file from the shelf' do
        client_as_super do |cs|
          cs.upload_file({DepotFile: '//depot/main/preview_test_remove/to_delete', 'Content': 'First state!'})
        end
        client_as_jdoe do |c|
          c.add_setting('HVE_PROJECTS_PATH', '//depot/main')

          begin
            p4 = p4_as_jdoe

            info = create_sync_project('//depot/main/preview_test_remove', c, p4)
            sync_root = info[0]
            device_client = info[1]

            file_path = File.join(sync_root, 'to_delete')
            File.delete(file_path)

            run_device_sync('//depot/main/preview_test_remove', sync_root, c, device_client)

            temp_client_as_super do |p4s|
              r = p4s.run_delete("//depot/main/preview_test_remove/to_delete")
              puts "remove from shelf delete: #{r}"
              r = p4s.run_submit('-d', "deleting to_delete")
              puts "remove from shelf submit: #{r}"
            end

            preview = c.preview_helix_sync_pending_change('preview_test_remove')

            puts "preview remove from shelf: #{preview}"

            expect(preview.length).to eq(1)

            expect(preview.first.depotFile).to eq('//depot/main/preview_test_remove/to_delete')
            expect(preview.first.action).to eq('delete')

            # Cleanup
            c.delete_helix_sync_pending_changelist('preview_test_remove')
            c.delete_helix_sync_device_client('preview_test_remove', 'test-machine')
            c.delete_helix_sync_shelf_client('preview_test_remove')

          ensure
            p4.disconnect if !p4.nil? and p4.connected?
            FileUtils.rmtree(sync_root) if sync_root && Dir.exist?(sync_root)
          end
        end
      end
    end

    # This is a scenario where user A edited a file that user B has deleted,
    # and user A submits the changes. The resolution is to "re-add".
    context 'readd' do
      it 'adds file content that the user has edited but someone else removed' do
        client_as_super do |cs|
          cs.upload_file({DepotFile: '//depot/main/preview_readd_test/README', 'Content': 'First state!'})
        end

        client_as_jdoe do |c|
          c.add_setting('HVE_PROJECTS_PATH', '//depot/main')

          begin
            p4 = p4_as_jdoe

            info = create_sync_project('//depot/main/preview_readd_test', c, p4)
            sync_root = info[0]
            device_client = info[1]

            readme_path = File.join(sync_root, 'README')
            IO.write(readme_path, 'Second state')

            run_device_sync('//depot/main/preview_readd_test', sync_root, c, device_client)

            temp_client_as_super do |p4s|
              r = p4s.run_delete('//depot/main/preview_readd_test/README')
              puts "readd delete: #{r}"
              r = p4s.run_submit('-d', 'deleting README')
              puts "readd submit: #{r}"
            end

            preview = c.preview_helix_sync_pending_change('preview_readd_test')

            expect(preview.first.depotFile).to eq('//depot/main/preview_readd_test/README')
            expect(preview.first.action).to eq('edit')

            # Cleanup
            c.delete_helix_sync_pending_changelist('preview_readd_test')
            c.delete_helix_sync_device_client('preview_readd_test', 'test-machine')
            c.delete_helix_sync_shelf_client('preview_readd_test')

          ensure
            p4.disconnect if !p4.nil? and p4.connected?
            FileUtils.rmtree(sync_root) if sync_root && Dir.exist?(sync_root)
          end
        end
      end
    end
  end

  context 'submit_helix_sync_pending_change' do
    new_file_name = nil

    context 'no resolves needed' do
      it 'creates a new file' do
        client_as_jdoe do |c|
          c.add_setting('HVE_PROJECTS_PATH', '//depot/main')

          begin
            p4 = p4_as_jdoe

            info = create_sync_project('//depot/main/synctest', c, p4)
            sync_root = info[0]
            device_client = info[1]

            new_file_name = create_new_file_under(sync_root)

            run_device_sync('//depot/main/synctest', sync_root, c, device_client)

            c.submit_helix_sync_pending_change('synctest')

            changes = p4.run_changes('-s', 'pending', '-u', 'jdoe')
            puts "no resolves needed changes: #{changes}"

            files = c.files('//depot/main/synctest')

            expect(files.map(&:depotFile)).to include("//depot/main/synctest/#{new_file_name}")

          ensure
            p4.disconnect if !p4.nil? and p4.connected?
            FileUtils.rmtree(sync_root) if sync_root && Dir.exist?(sync_root)
          end
        end
      end
    end

    # This is a scenario where we've deleted a file on a shelf, that was also
    # already deleted by someone else. So the "resolution" is to just remove
    # the file from the shelf, no further action is needed.
    context 'remove from shelf' do
      # This test depends on the file added above
      it 'just removes the file from the shelf' do
        client_as_jdoe do |c|
          c.add_setting('HVE_PROJECTS_PATH', '//depot/main')

          begin
            p4 = p4_as_jdoe

            info = create_sync_project('//depot/main/synctest', c, p4)
            sync_root = info[0]
            device_client = info[1]

            new_file_path = File.join(sync_root, new_file_name)
            File.delete(new_file_path)

            run_device_sync('//depot/main/synctest', sync_root, c, device_client)

            temp_client_as_super do |p4s|
              r = p4s.run_delete("//depot/main/synctest/#{new_file_name}")
              puts "remove from shelf delete: #{r}"
              r = p4s.run_submit('-d', "deleting #{new_file_name}")
              puts "remove from shelf submit: #{r}"
            end

            c.submit_helix_sync_pending_change('synctest')

          ensure
            p4.disconnect if !p4.nil? and p4.connected?
            FileUtils.rmtree(sync_root) if sync_root && Dir.exist?(sync_root)
          end
        end
      end
    end

    # This is a scenario where user A edited a file that user B has deleted,
    # and user A submits the changes. The resolution is to "re-add".
    context 'readd' do
      it 'adds file content that the user has edited but someone else removed' do
        client_as_super do |cs|
          cs.upload_file({DepotFile: '//depot/main/readd_test/README', 'Content': 'First state!'})
        end

        client_as_jdoe do |c|
          c.add_setting('HVE_PROJECTS_PATH', '//depot/main')

          begin
            p4 = p4_as_jdoe

            info = create_sync_project('//depot/main/readd_test', c, p4)
            sync_root = info[0]
            device_client = info[1]

            readme_path = File.join(sync_root, 'README')
            IO.write(readme_path, 'Second state')

            run_device_sync('//depot/main/readd_test', sync_root, c, device_client)

            temp_client_as_super do |p4s|
              r = p4s.run_delete('//depot/main/readd_test/README')
              puts "readd delete: #{r}"
              r = p4s.run_submit('-d', 'deleting README')
              puts "readd submit: #{r}"
            end

            c.submit_helix_sync_pending_change('readd_test')

            files = c.files('//depot/main/readd_test')

            shelf_client = c.create_helix_sync_shelf_client('readd_test').client
            changes = p4.run_changes('-s', 'pending', '-u', 'jdoe', '-c', shelf_client)

            puts "post readd changes: #{changes}"

            expect(changes.empty?).to be(true)

            file = c.file('//depot/main/readd_test/README')

            expect(file.Content).to eq('Second state')

          ensure
            p4.disconnect if !p4.nil? and p4.connected?
            FileUtils.rmtree(sync_root) if sync_root && Dir.exist?(sync_root)
          end
        end
      end
    end

    context 'lock' do
      it 'renames a locked file from the submit' do
        randstr = (0...8).map { (65 + rand(26)).chr }.join
        lock_file_name = "test_#{randstr}"

        project_id = "lock_test_#{randstr}"

        client_as_super do |cs|
          cs.upload_file({DepotFile: "//depot/main/#{project_id}/#{lock_file_name}", 'Content': 'Current State'})
        end

        client_as_jdoe do |c|
          c.add_setting('HVE_PROJECTS_PATH', '//depot/main')

          begin
            p4 = p4_as_jdoe

            info = create_sync_project("//depot/main/#{project_id}", c, p4)
            sync_root = info[0]
            device_client = info[1]

            new_file_name = create_new_file_under(sync_root)

            IO.write("#{sync_root}/#{lock_file_name}", 'Edited state')

            # The server seems to run into problems if we run device sync too fast
            # I'm adding 1-second sleeps to ensure there's at least a wee bit
            # of time between operations. Otherwise fstat just doesn't report
            # the locks are actually locks
            sleep(1)

            run_device_sync("//depot/main/#{project_id}", sync_root, c, device_client)

            sleep(1)

            temp_client_as_super do |p4s|
              edit_results = p4s.run_edit("//depot/main/#{project_id}/#{lock_file_name}")
              puts "edit_results #{edit_results}"
              lock_results = p4s.run_lock("//depot/main/#{project_id}/#{lock_file_name}")
              puts "lock_results #{lock_results}"
              true
            end

            sleep(1)

            pending = c.fetch_helix_sync_pending_changelist(project_id).change
            puts "pending change #{pending}"

            c.submit_helix_sync_pending_change(project_id)

            changes = p4.run_changes('-s', 'pending', '-u', 'jdoe')
            puts "no resolves needed changes: #{changes}"

            files = c.files("//depot/main/#{project_id}")

            expect(files.map(&:depotFile)).to include("//depot/main/#{project_id}/#{lock_file_name}.#{pending}.locked")

            lock_file = c.file("//depot/main/#{project_id}/#{lock_file_name}")

            expect(lock_file.Content).to eq('Current State')

            renamed_file = c.file("//depot/main/#{project_id}/#{lock_file_name}.#{pending}.locked")

            expect(renamed_file.Content).to eq('Edited state')

          ensure
            p4.disconnect if !p4.nil? and p4.connected?
            FileUtils.rmtree(sync_root) if sync_root && Dir.exist?(sync_root)
          end
        end
      end
    end
  end
end
